
import math
import os
import time
import logging
import warnings
from typing import Optional, Callable, List, Tuple, Dict

import numpy as np
import torch
from torch import nn, optim

import dgl
from dgl import DGLGraph
from dgl.dataloading import GraphDataLoader, GraphCollator

# TODO: sklearn metrics functions
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
from sklearn.exceptions import UndefinedMetricWarning

from training_args import TrainingArguments

from optimization import get_scheduler

from callbacks import (
    TrainerState,
    TrainerControl,
    TrainerCallback,
    CallbackHandler,
    DefaultFlowCallback,
    EarlyStoppingCallback,
    PrinterCallback,
    ProgressCallback,
)

from trainer_utils import set_seed, get_parameter_names, speed_metrics

logger = logging.getLogger(__name__)

warnings.filterwarnings('ignore', category=UndefinedMetricWarning)

# Names of files used for checkpointing
CONFIG_NAME = 'config.json'
WEIGHTS_NAME = 'pytorch_model.bin'
TRAINING_ARGS_NAME = 'training_args.bin'
TRAINER_STATE_NAME = 'trainer_state.json'
OPTIMIZER_NAME = 'optimizer.pt'
SCHEDULER_NAME = 'scheduler.pt'
PREFIX_CHECKPOINT_DIR = 'checkpoint'

DEFAULT_CALLBACKS = [DefaultFlowCallback]
DEFAULT_PROGRESS_CALLBACK = ProgressCallback

# TODO: Saving model, optimizer, scheduler states
# TODO: Checkpoint limit. Delete older checkpoints
# TODO: Resuming training from checkpoint
# TODO: Half-precision support
# TODO: Distributed data parallel support
"""
https://github.com/huggingface/transformers/tree/v4.15.0/src/transformers
https://github.com/huggingface/transformers/blob/v4.15.0/src/transformers/trainer.py
"""


class Trainer:
    def __init__(
            self,
            model: nn.Module = None,
            loss_module: nn.Module = None,
            args: TrainingArguments = None,
            collate_fn: Optional[Callable] = None,
            train_dataset: Optional[Tuple[List[DGLGraph], torch.Tensor]] = (None, None),
            eval_dataset: Optional[Tuple[List[DGLGraph], torch.Tensor]] = (None, None),
            callbacks: Optional[List[TrainerCallback]] = None,
            optimizers: Tuple[optim.Optimizer, optim.lr_scheduler.LambdaLR] = (None, None),
    ):
        if args is None:
            output_dir = 'tmp_trainer'
            logger.info(f'No `TrainingArguments` passed, using `output_dir={output_dir}.')
            args = TrainingArguments(output_dir=output_dir)
        self.args = args

        set_seed(self.args.seed)
        self.is_in_train = False

        log_level = args.get_log_level()
        logger.setLevel(log_level)

        if model is None:
            raise RuntimeError('`Trainer` requires a `model`.')

        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset

        default_collate_fn = GraphCollator
        self.collate_fn = collate_fn if collate_fn is not None else default_collate_fn

        self.model = self._move_model_to_device(model, args.device)

        if loss_module is not None:
            loss_module = self._move_model_to_device(loss_module, args.device)
        self.loss_module = loss_module

        self.optimizer, self.lr_scheduler = optimizers

        default_callbacks = DEFAULT_CALLBACKS
        callbacks = default_callbacks if callbacks is None else default_callbacks + callbacks
        self.callback_handler = CallbackHandler(
            callbacks, self.model, self.optimizer, self.lr_scheduler
        )
        self.add_callback(PrinterCallback if self.args.disable_tqdm else DEFAULT_PROGRESS_CALLBACK)
        if args.do_early_stopping:
            self.add_callback(EarlyStoppingCallback)

        self._loggers_initialized = False

        os.makedirs(self.args.output_dir, exist_ok=True)

        if self.collate_fn is not None and not callable(getattr(self.collate_fn, 'collate', None)):
            raise ValueError('The `collate_fn` should have a callable "collate" function.')

        self.state = TrainerState()
        self.control = TrainerControl()

    @staticmethod
    def _move_model_to_device(model, device):
        return model.to(device)

    @staticmethod
    def num_examples(dataloader):
        if dgl.__version__.startswith('0.7'):
            return len(dataloader.dataloader.dataset)
        else:
            return len(dataloader.dataset)

    def add_callback(self, callback):
        self.callback_handler.add_callback(callback)

    def pop_callback(self, callback):
        return self.callback_handler.pop_callback(callback)

    def remove_callback(self, callback):
        self.callback_handler.remove_callback(callback)

    def get_train_dataloader(self) -> GraphDataLoader:
        if self.train_dataset is None:
            raise ValueError('Trainer: training requires a train_dataset.')

        train_dataset = self.train_dataset

        return GraphDataLoader(
            dataset=train_dataset,
            batch_size=self.args.train_batch_size,
            collate_fn=None,
            num_workers=self.args.dataloader_num_workers,
            pin_memory=self.args.dataloader_pin_memory,
            drop_last=self.args.dataloader_drop_last,
            shuffle=True,
        )

    def get_eval_dataloader(
            self, eval_dataset: Optional[Tuple[List[DGLGraph], torch.Tensor]] = None
    ) -> GraphDataLoader:

        if eval_dataset is None and self.eval_dataset is None:
            raise ValueError('Trainer: evaluation requires an eval_dataset.')
        eval_dataset = eval_dataset if eval_dataset is not None else self.eval_dataset

        return GraphDataLoader(
            dataset=eval_dataset,
            batch_size=self.args.eval_batch_size,
            collate_fn=None,
            num_workers=self.args.dataloader_num_workers,
            pin_memory=self.args.dataloader_pin_memory,
            drop_last=self.args.dataloader_drop_last,
            shuffle=False,
        )

    def get_test_dataloader(self, test_dataset: Tuple[List[DGLGraph], torch.Tensor] = None) -> GraphDataLoader:
        return GraphDataLoader(
            dataset=test_dataset,
            batch_size=self.args.eval_batch_size,
            collate_fn=None,
            num_workers=self.args.dataloader_num_workers,
            pin_memory=self.args.dataloader_pin_memory,
            drop_last=self.args.dataloader_drop_last,
            shuffle=False,
        )

    def create_optimizer(self):
        if self.optimizer is None:
            optimizer_kwargs = {
                'betas': (self.args.adam_beta1, self.args.adam_beta2),
                'eps': self.args.adam_epsilon,
                'lr': self.args.learning_rate
            }

            model_decay_parameters = get_parameter_names(self.model, [nn.LayerNorm])
            model_decay_parameters = [name for name in model_decay_parameters if 'bias' not in name]
            optimizer_grouped_parameters = [
                {
                    'params': [p for n, p in self.model.named_parameters() if n in model_decay_parameters],
                    'weight_decay': self.args.weight_decay,
                },
                {
                    'params': [p for n, p in self.model.named_parameters() if n not in model_decay_parameters],
                    'weight_decay': 0.0,
                },
            ]

            if self.args.trainable_loss:
                loss_decay_parameters = get_parameter_names(self.loss_module, [nn.LayerNorm])
                loss_decay_parameters = [name for name in loss_decay_parameters if 'bias' not in name]

                optimizer_grouped_parameters.append({
                    'params': [p for n, p in self.loss_module.named_parameters() if n in loss_decay_parameters],
                    'weight_decay': self.args.weight_decay,
                })
                optimizer_grouped_parameters.append({
                    'params': [p for n, p in self.loss_module.named_parameters() if n not in loss_decay_parameters],
                    'weight_decay': 0.0,
                })

            self.optimizer = optim.AdamW(optimizer_grouped_parameters, **optimizer_kwargs)

        return self.optimizer

    def create_optimizer_and_scheduler(self, num_training_steps: int):
        self.create_optimizer()
        self.create_scheduler(num_training_steps=num_training_steps, optimizer=self.optimizer)

    def create_scheduler(self, num_training_steps: int, optimizer: optim.Optimizer = None):
        if self.lr_scheduler is None:
            self.lr_scheduler = get_scheduler(
                self.args.lr_scheduler_type,
                optimizer=self.optimizer if optimizer is None else optimizer,
                num_warmup_steps=self.args.get_warmup_steps(num_training_steps),
                num_training_steps=num_training_steps,
            )
        return self.lr_scheduler

    def train(
            self,
            **kwargs
    ):
        args = self.args

        self.is_in_train = True

        self.model = self._move_model_to_device(self.model, args.device)
        if self.loss_module is not None:
            self.loss_module = self._move_model_to_device(self.loss_module, args.device)

        train_dataloader = self.get_train_dataloader()

        # Set up training control variables
        total_train_batch_size = args.train_batch_size * args.gradient_accumulation_steps
        num_update_steps_per_epoch = max(len(train_dataloader) // args.gradient_accumulation_steps, 1)

        max_steps = math.ceil(args.num_train_epochs * num_update_steps_per_epoch)
        num_train_epochs = math.ceil(args.num_train_epochs)
        num_train_samples = len(self.train_dataset) * args.num_train_epochs

        self.create_optimizer_and_scheduler(num_training_steps=max_steps)

        self.state = TrainerState()

        model = self._move_model_to_device(self.model, args.device)
        loss_module = None if self.loss_module is None else self._move_model_to_device(self.loss_module, args.device)

        # Train
        num_examples = self.num_examples(train_dataloader)

        logger.info('***** Running Training *****')
        logger.info(f'  Num examples = {num_examples}')
        logger.info(f'  Num epochs = {num_train_epochs}')
        logger.info(f'  Batch size = {args.train_batch_size}')
        logger.info(f'  Gradient accumulation steps = {args.gradient_accumulation_steps}')
        logger.info(f'  Total optimization steps = {max_steps}')

        self.state.epoch = 0
        start_time = time.time()
        epochs_trained = 0

        # Update references
        self.callback_handler.model = self.model
        self.callback_handler.optimizer = self.optimizer
        self.callback_handler.lr_scheduler = self.lr_scheduler
        self.callback_handler.train_dataloader = train_dataloader
        # These should be the same if the state has been saved but in case the training arguments changes,
        # it's safer to set this after the load.
        self.state.max_steps = max_steps
        self.state.num_train_epochs = num_train_epochs

        tr_loss = torch.tensor(0.0, device=args.device)
        self._total_loss_scalar = 0.0
        self._global_step_last_logged = self.state.global_step

        self.optimizer.zero_grad()

        self.control = self.callback_handler.on_train_begin(args, self.state, self.control)

        metrics = {}
        for epoch in range(epochs_trained, num_train_epochs):
            epoch_iterator = train_dataloader

            steps_in_epoch = len(epoch_iterator)

            self.control = self.callback_handler.on_epoch_begin(args, self.state, self.control)

            step = -1
            for step, inputs in enumerate(epoch_iterator):
                if step % args.gradient_accumulation_steps == 0:
                    self.control = self.callback_handler.on_step_begin(args, self.state, self.control)

                tr_loss_step = self.training_step(model, loss_module, inputs)

                if (
                        args.logging_nan_inf_filter
                        and (torch.isnan(tr_loss_step) or torch.isinf(tr_loss_step))
                ):
                    # If loss is nan or inf, simply add the average of previous logged losses
                    tr_loss += tr_loss / (1 + self.state.global_step - self._global_step_last_logged)
                else:
                    tr_loss += tr_loss_step

                if (step + 1) % args.gradient_accumulation_steps == 0 or (
                        # Last step in the epoch, but step is always smaller than gradient_accumulation_steps
                        args.gradient_accumulation_steps >= steps_in_epoch == (step + 1)
                ):
                    # Gradient clipping
                    if args.max_grad_norm is not None and args.max_grad_norm > 0:
                        if hasattr(self.optimizer, 'clip_grad_norm'):
                            self.optimizer.clip_grad_norm(args.max_grad_norm)
                        elif hasattr(model, 'clip_grad_norm_'):
                            model.clip_grad_norm_(args.max_grad_norm)
                        else:
                            nn.utils.clip_grad_norm_(
                                model.parameters(),
                                args.max_grad_norm
                            )

                    # Optimizer step
                    self.optimizer.step()
                    self.lr_scheduler.step()

                    self.optimizer.zero_grad()

                    self.state.global_step += 1
                    self.state.epoch = epoch + (step + 1) / steps_in_epoch
                    self.control = self.callback_handler.on_step_end(
                        args, self.state, self.control, loss=tr_loss.item()
                    )

                    tmp_metrics = self._maybe_log_save_evaluate(tr_loss, model, epoch)
                    if tmp_metrics:
                        metrics[f'epoch_{epoch:02d}_{step:03d}'] = tmp_metrics

                else:
                    self.control = self.callback_handler.on_substep_end(args, self.state, self.control)

                if self.control.should_epoch_stop or self.control.should_training_stop:
                    break

            if step < 0:
                logger.warning(
                    f'There seems to be not a single sample in your epoch_iterator, stopping training at step '
                    f'{self.state.global_step}! This is expected if you\'re using an IterableDataset and set '
                    f'num_steps ({max_steps}) higher than the number of available samples.'
                )
                self.control.should_training_stop = True

            self.control = self.callback_handler.on_epoch_end(args, self.state, self.control)
            tmp_metrics = self._maybe_log_save_evaluate(tr_loss, model, epoch)
            if tmp_metrics:
                metrics[f'epoch_{epoch:02d}'] = tmp_metrics

            if self.control.should_training_stop:
                break

        logger.info(f'\n\nTraining completed.')
        if args.load_best_model_at_end and self.state.best_model_checkpoint is not None:
            logger.info(
                f'Loading best model from {self.state.best_model_checkpoint} (score: {self.state.best_metric}).'
            )

            best_model_path = os.path.join(self.state.best_model_checkpoint, WEIGHTS_NAME)
            if os.path.exists(best_model_path):
                state_dict = torch.load(best_model_path, map_location='cpu')
                self._load_state_dict_in_model(state_dict)
            else:
                logger.warning(f'Could not locate the best model at {best_model_path}')

        self._total_loss_scalar += tr_loss.item()
        train_loss = self._total_loss_scalar / self.state.global_step

        # metrics = speed_metrics('train', start_time, num_samples=num_train_samples, num_steps=self.state.max_steps)
        metrics['train_loss'] = train_loss

        self.is_in_train = False

        self.log(metrics)

        self.control = self.callback_handler.on_train_end(args, self.state, self.control)

        return self.state.global_step, train_loss, metrics

    def training_step(self, model: nn.Module, loss_module: nn.Module, inputs: Tuple[DGLGraph, torch.Tensor]):
        model.train()
        inputs = self._prepare_inputs(inputs)

        loss = self.compute_loss(model, loss_module, inputs)

        if self.args.gradient_accumulation_steps > 1:
            loss /= self.args.gradient_accumulation_steps

        loss.backward()

        return loss.detach()

    def compute_loss(
            self,
            model: nn.Module,
            loss_module: nn.Module,
            inputs: Tuple[DGLGraph, torch.Tensor],
            return_outputs: bool = False
    ):
        batched_graph, labels = inputs
        outputs = model(batched_graph, batched_graph.ndata.pop('feat').float())

        outputs = outputs.squeeze()
        labels = labels.squeeze()

        loss = loss_module(outputs, labels)

        return (loss, outputs) if return_outputs else loss

    def _prepare_inputs(self, inputs: Tuple[DGLGraph, torch.Tensor]):
        batched_graph, labels = inputs
        return batched_graph.to(self.args.device), labels.squeeze().to(self.args.device)

    def _load_state_dict_in_model(self, state_dict):
        load_result = self.model.load_state_dict(state_dict, strict=False)

        if len(load_result.missing_keys) != 0:
            pass

    def load_model(self, checkpoint_dir: str, model=None):
        if self.model is None and model is None:
            raise ValueError('No model provided.')
        elif model:
            self.model = model

        if not os.path.isfile(os.path.join(checkpoint_dir, WEIGHTS_NAME)):
            raise ValueError(f'Cannot find a valid checkpoint at {checkpoint_dir}')

        state_dict = torch.load(os.path.join(checkpoint_dir, WEIGHTS_NAME), map_location='cpu')
        self._load_state_dict_in_model(state_dict)
        del state_dict  # Free memory

    def log(self, logs: Dict[str, float]) -> None:
        if self.state.epoch is not None:
            logs['epoch'] = round(self.state.epoch, 2)

        output = {**logs, **{'step': self.state.global_step}}
        self.state.log_history.append(output)
        self.control = self.callback_handler.on_log(self.args, self.state, self.control, logs=logs)

    def _maybe_log_save_evaluate(self, tr_loss, model, epoch):
        if self.control.should_log:
            logs: Dict[str, float] = {}

            tr_loss_scalar = tr_loss.mean().item()
            tr_loss -= tr_loss

            logs['loss'] = round(tr_loss_scalar / (self.state.global_step - self._global_step_last_logged), 4)
            logs['learning_rate'] = self._get_learning_rate()

            self._total_loss_scalar += tr_loss_scalar
            self._global_step_last_logged = self.state.global_step

            self.log(logs)

        metrics = None
        if self.control.should_evaluate:
            metrics = self.evaluate()
            self.control = self.callback_handler.on_evaluate(self.args, self.state, self.control, metrics=metrics)

        if self.control.should_save:
            self._save_checkpoint(model, metrics=metrics)
            self.control = self.callback_handler.on_save(self.args, self.state, self.control)

        return metrics

    def _get_learning_rate(self):
        return self.lr_scheduler.get_last_lr()[0]

    def _save_checkpoint(self, model, metrics=None):
        checkpoint_folder = f'{PREFIX_CHECKPOINT_DIR}-{self.state.global_step}'
        output_dir = os.path.join(self.args.output_dir, checkpoint_folder)
        os.makedirs(output_dir, exist_ok=True)

        self.save_model(output_dir)

        if metrics is not None and self.args.metric_for_best_model is not None:
            metric_to_check = self.args.metric_for_best_model
            if not metric_to_check.startswith('eval_'):
                metric_to_check = f'eval_{metric_to_check}'
            metric_value = metrics[metric_to_check]

            operator = np.greater if self.args.greater_is_better else np.less
            if (
                    self.state.best_metric is None
                    or self.state.best_model_checkpoint is None
                    or operator(metric_value, self.state.best_metric)
            ):
                self.state.best_metric = metric_value
                self.state.best_model_checkpoint = output_dir

        self.state.save_to_json(os.path.join(output_dir, TRAINER_STATE_NAME))

    def save_model(self, output_dir: Optional[str] = None):
        if output_dir is None:
            output_dir = self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)

        torch.save(self.model.state_dict(), os.path.join(output_dir, WEIGHTS_NAME))
        torch.save(self.args, os.path.join(output_dir, TRAINING_ARGS_NAME))

    def evaluate(
            self,
            eval_dataset: Optional[Tuple[List[DGLGraph], torch.Tensor]] = None,
            metric_key_prefix: str = 'eval',
    ) -> Dict[str, float]:

        eval_dataloader = self.get_eval_dataloader(eval_dataset)
        num_samples = self.num_examples(eval_dataloader)

        start_time = time.time()
        metrics = self.evaluation_loop(
            eval_dataloader,
            description='Evaluation',
            prediction_loss_only=None,
            metric_key_prefix=metric_key_prefix
        )

        # metrics.update(
        #     speed_metrics(
        #         metric_key_prefix,
        #         start_time,
        #         num_samples=num_samples,
        #         num_steps=math.ceil(num_samples / self.args.eval_batch_size),
        #     )
        # )

        self.log(metrics)

        return metrics

    # TODO: Temp hard-code evaluation metrics
    def compute_metrics(self, preds: torch.Tensor, labels: torch.Tensor, threshold: float = 0.5):
        is_multiclass = labels.max().item() > 1
        if is_multiclass:
            preds = torch.argmax(preds, dim=-1)
            probs = preds.tolist()  # Predicted class not raw probs
        else:
            probs = preds.tolist()
            preds = (preds > threshold).float()

        return {
            'accuracy': accuracy_score(preds, labels),
            'precision': precision_score(preds, labels, average='micro' if is_multiclass else 'binary'),
            'recall': recall_score(preds, labels, average='micro' if is_multiclass else 'binary'),
            'F1 micro': f1_score(preds, labels, average='micro'),
            'F1 macro': f1_score(preds, labels, average='macro'),
            'probs': probs,
            'labels': labels.tolist(),
        }

    def evaluation_loop(
            self,
            dataloader: GraphDataLoader,
            description: str,
            prediction_loss_only: Optional[bool] = None,
            metric_key_prefix: str = 'eval',
    ):
        args = self.args

        prediction_loss_only = prediction_loss_only if prediction_loss_only is not None else args.prediction_loss_only

        logger.info(f'***** Running {description} *****')
        logger.info(f'  Num examples = {self.num_examples(dataloader)}')
        logger.info(f'  Batch size = {args.eval_batch_size}')

        model = self.model
        model.eval()

        self.callback_handler.eval_dataloader = dataloader

        all_losses, all_preds, all_labels = None, None, None
        for step, inputs in enumerate(dataloader):
            losses, logits, labels = self.prediction_step(model, inputs, prediction_loss_only)

            self.control = self.callback_handler.on_prediction_step(args, self.state, self.control)

            losses = losses.unsqueeze(-1)

            try:
                all_losses = losses if all_losses is None else torch.cat((all_losses, losses), dim=0)
                all_preds = logits if all_preds is None else torch.cat((all_preds, logits), dim=0)
                all_labels = labels if all_labels is None else torch.cat((all_labels, labels), dim=0)
            except RuntimeError:
                print(logits.shape)
                print(all_preds if all_preds is None else all_preds.shape)

        all_losses = all_losses.cpu()
        all_preds = all_preds.cpu()
        all_labels = all_labels.cpu()

        metrics = self.compute_metrics(all_preds, all_labels)

        # TODO: Remove tensor/np types for JSON serialization

        if all_losses is not None:
            metrics[f'{metric_key_prefix}_loss'] = all_losses.mean().item()

        metrics_with_prefix = {}
        for key in metrics:
            val = metrics[key]
            if not key.startswith(f'{metric_key_prefix}_'):
                metrics_with_prefix[f'{metric_key_prefix}_{key}'] = val
            else:
                metrics_with_prefix[key] = val

        return metrics_with_prefix

    def prediction_step(
            self,
            model: nn.Module,
            inputs: Tuple[DGLGraph, torch.Tensor],
            prediction_loss_only: bool,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        inputs = self._prepare_inputs(inputs)

        labels = inputs[1].detach()

        with torch.no_grad():
            loss, logits = self.compute_loss(model, self.loss_module, inputs, return_outputs=True)
        loss = loss.mean().detach()

        if prediction_loss_only:
            return loss, None, None

        preds = torch.sigmoid(logits.detach())
        # _, preds = logits.detach().max(dim=-1)

        return loss, preds, labels




import json
import logging
from dataclasses import dataclass, asdict
from typing import Optional, List, Dict

import numpy as np
from tqdm.auto import tqdm

from training_args import TrainingArguments
from trainer_utils import IntervalStrategy


"""
Trainer Callbacks adapted from HuggingFace:
https://github.com/huggingface/transformers/blob/v4.15.0/src/transformers/trainer_callback.py
"""

logger = logging.getLogger(__name__)


@dataclass
class TrainerState:
    epoch: Optional[float] = None
    global_step: int = 0
    max_steps: int = 0
    num_train_epochs: int = 0
    log_history: List[Dict[str, float]] = None
    best_metric: Optional[float] = None
    best_model_checkpoint: Optional[str] = None

    def __post_init__(self):
        if self.log_history is None:
            self.log_history = []

    def save_to_json(self, json_path: str):
        json_string = json.dumps(asdict(self), indent=2, sort_keys=True) + '\n'
        with open(json_path, 'w', encoding='utf-8')as f:
            f.write(json_string)

    @classmethod
    def load_from_json(cls, json_path:str):
        with open(json_path, 'r', encoding='utf-8') as f:
            text = f.read()
        return cls(**json.loads(text))


@dataclass
class TrainerControl:
    should_training_stop: bool = False
    should_epoch_stop: bool = False
    should_save: bool = False
    should_evaluate: bool = False
    should_log: bool = False

    def _new_training(self):
        """Internal method that resets the variable for a new training."""
        self.should_training_stop = False

    def _new_epoch(self):
        """Internal method that resets the variable for a new epoch."""
        self.should_epoch_stop = False

    def _new_step(self):
        """Internal method that resets the variable for a new step."""
        self.should_save = False
        self.should_evaluate = False
        self.should_log = False


class TrainerCallback:
    def on_init_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the end of the initialization of the [`Trainer`].
        """
        pass

    def on_train_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the beginning of training.
        """
        pass

    def on_train_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the end of training.
        """
        pass

    def on_epoch_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the beginning of an epoch.
        """
        pass

    def on_epoch_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the end of an epoch.
        """
        pass

    def on_step_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the beginning of a training step. If using gradient accumulation, one training step might take
        several inputs.
        """
        pass

    def on_substep_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the end of a substep during gradient accumulation.
        """
        pass

    def on_step_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the end of a training step. If using gradient accumulation, one training step might take
        several inputs.
        """
        pass

    def on_evaluate(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called after an evaluation phase.
        """
        pass

    def on_save(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called after a checkpoint save.
        """
        pass

    def on_log(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called after logging the last logs.
        """
        pass

    def on_prediction_step(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called after a prediction step.
        """
        pass


class CallbackHandler(TrainerCallback):
    """Internal class that just calls the list of callbacks in order"""
    def __init__(self, callbacks, model, optimizer, lr_scheduler):
        self.callbacks = []
        for cb in callbacks:
            self.add_callback(cb)
        self.model = model
        self.optimizer = optimizer
        self.scheduler = lr_scheduler
        self.train_dataloader = None
        self.eval_dataloader = None

        if not any(isinstance(cb, DefaultFlowCallback) for cb in self.callbacks):
            logger.warning(
                'The Trainer will not work properly without a `DefaultFlowCallback` in its callbacks.\n'
                'You should add one before training with `trainer.add_callback(DefaultFlowCallback).\n'
                f'The current list of callbacks is:\n'
                f'{self.callback_list}'
            )

    def add_callback(self, callback):
        cb = callback() if isinstance(callback, type) else callback
        cb_class = callback if isinstance(callback, type) else callback.__class__
        if cb_class in [c.__class__ for c in self.callbacks]:
            logger.warning(
                f'You are adding a {cb_class} to the callbacks of this Trainer, but there is already one.\n'
                f'The current list of callbacks is:\n'
                f'{self.callback_list}'
            )
        self.callbacks.append(cb)

    def pop_callback(self, callback):
        if isinstance(callback, type):
            for cb in self.callbacks:
                if isinstance(cb, callback):
                    self.callbacks.remove(cb)
                    return cb
        else:
            for cb in self.callbacks:
                if cb == callback:
                    self.callbacks.remove(cb)
                    return cb

    def remove_callback(self, callback):
        if isinstance(callback, type):
            for cb in self.callbacks:
                if isinstance(cb, callback):
                    self.callbacks.remove(cb)
                    return
        else:
            self.callbacks.remove(callback)

    @property
    def callback_list(self):
        return '\n'.join(cb.__class__.__name__ for cb in self.callbacks)

    def call_event(self, event: str, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        for callback in self.callbacks:
            result = getattr(callback, event)(
                args,
                state,
                control,
                model=self.model,
                optimizer=self.optimizer,
                lr_scheduler=self.scheduler,
                train_dataloader=self.train_dataloader,
                eval_dataloader=self.eval_dataloader,
                **kwargs,
            )
            # A Callback can skip the return of `control` if it doesn't change it.
            if result is not None:
                control = result
        return control

    def on_init_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        return self.call_event('on_init_end', args, state, control)

    def on_train_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        control.should_training_stop = False
        return self.call_event('on_train_begin', args, state, control)

    def on_train_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        return self.call_event('on_train_end', args, state, control)

    def on_epoch_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        control.should_epoch_stop = False
        return self.call_event('on_epoch_begin', args, state, control)

    def on_epoch_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        return self.call_event('on_epoch_end', args, state, control)

    def on_step_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        control.should_log = False
        control.should_evaluate = False
        control.should_save = False
        return self.call_event('on_step_begin', args, state, control)

    def on_substep_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        return self.call_event('on_substep_end', args, state, control)

    def on_step_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        return self.call_event('on_step_end', args, state, control)

    def on_evaluate(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        control.should_evaluate = False
        return self.call_event('on_evaluate', args, state, control, **kwargs)

    def on_save(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        control.should_save = False
        return self.call_event('on_save', args, state, control)

    def on_log(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        control.should_log = False
        return self.call_event('on_log', args, state, control, **kwargs)

    def on_prediction_step(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        return self.call_event('on_prediction_step', args, state, control)


class DefaultFlowCallback(TrainerCallback):
    def on_epoch_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        # Log
        if args.logging_strategy == IntervalStrategy.EPOCH:
            control.should_log = True

        # Evaluate
        if args.evaluation_strategy == IntervalStrategy.EPOCH:
            control.should_evaluate = True

        # Save
        if args.save_strategy == IntervalStrategy.EPOCH:
            control.should_save = True

        return control


class ProgressCallback(TrainerCallback):
    """
    A [`TrainerCallback`] that displays the progress of training or evaluation.
    """

    def __init__(self):
        self.training_bar = None
        self.prediction_bar = None
        self.current_step = None

    def on_train_begin(self, args, state, control, **kwargs):
        self.training_bar = tqdm(total=state.max_steps)
        self.current_step = 0

    def on_step_end(self, args, state, control, loss=None, **kwargs):
        self.training_bar.update(state.global_step - self.current_step)
        self.current_step = state.global_step
        if loss:
            self.training_bar.set_postfix({
                'loss': f'{loss:.4f}'
            })

    def on_prediction_step(self, args, state, control, eval_dataloader=None, **kwargs):
        if self.prediction_bar is None:
            self.prediction_bar = tqdm(total=len(eval_dataloader), leave=self.training_bar is None)
        self.prediction_bar.update(1)

    def on_evaluate(self, args, state, control, **kwargs):
        if self.prediction_bar is not None:
            self.prediction_bar.close()
        self.prediction_bar = None

    def on_log(self, args, state, control, logs=None, **kwargs):
        pass

    def on_train_end(self, args, state, control, **kwargs):
        self.training_bar.close()
        self.training_bar = None


class PrinterCallback(TrainerCallback):
    """
    A bare [`TrainerCallback`] that just prints the logs.
    """

    def on_log(self, args, state, control, logs=None, **kwargs):
        print(logs)


class EarlyStoppingCallback(TrainerCallback):
    """
    A [`TrainerCallback`] that handles early stopping.
    Args:
        early_stopping_patience: (`int`):
            Use with `metric_for_best_model` to stop training when the specified metric worsens for
            `early_stopping_patience` evaluation calls.
    This callback depends on [`TrainingArguments`] argument *load_best_model_at_end* functionality
    to set best_metric in [`TrainerState`].
    """

    def __init__(self, early_stopping_patience: int = 5):
        self.early_stopping_patience = early_stopping_patience
        self.early_stopping_patience_counter = 0

    def check_metric_value(self, args, state, metric_value):
        operator = np.greater if args.greater_is_better else np.less
        if state.best_metric is None or operator(metric_value, state.best_metric):
            self.early_stopping_patience_counter = 0
        else:
            self.early_stopping_patience_counter += 1

    def on_train_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        assert args.load_best_model_at_end, 'Early stopping requires load_best_model_at_end = True.'
        assert args.metric_for_best_model is not None, 'Early stopping requires metric_for_best_model is defined.'
        assert args.evaluation_strategy != IntervalStrategy.NO, 'Early stopping requires IntervalStrategy of steps or epochs'

    def on_evaluate(
            self, args: TrainingArguments, state: TrainerState, control: TrainerControl, metrics: dict = None, **kwargs
    ):
        metric_to_check = args.metric_for_best_model
        if not metric_to_check.startswith('eval_'):
            metric_to_check = f'eval_{metric_to_check}'
        metric_value = metrics.get(metric_to_check)

        if metric_value is None:
            logger.warning(
                f'Early stopping requires metric_for_best_model, '
                f'but did not find {metric_to_check} so early stopping is disabled.'
            )
            return

        self.check_metric_value(args, state, metric_value)
        if self.early_stopping_patience_counter > self.early_stopping_patience:
            control.should_training_stop = True




from enum import Enum
import random
import time

import numpy as np
import torch


class IntervalStrategy(Enum):
    NO = 'no'
    STEPS = 'steps'
    EPOCH = 'epoch'


def set_seed(seed: int):
    random.seed(seed)
    np.random.seed(seed)

    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def get_parameter_names(model, forbidden_layer_types):
    """
    Returns the names of the model parameters that are not inside a forbidden layer.
    """
    result = []
    for name, child in model.named_children():
        result += [
            f'{name}.{n}'
            for n in get_parameter_names(child, forbidden_layer_types)
            if not isinstance(child, tuple(forbidden_layer_types))
        ]
    # Add model specific parameters (defined with nn.Parameter) since they are not in any child.
    result += list(model._parameters.keys())

    return result


def speed_metrics(split, start_time, num_samples=None, num_steps=None):
    """
    Measure and return speed performance metrics.
    This function requires a time snapshot `start_time` before the operation to be measured starts and this function
    should be run immediately after the operation to be measured has completed.
    Args:
    - split: name to prefix metric (like train, eval, test...)
    - start_time: operation start time
    - num_samples: number of samples processed
    """
    runtime = time.time() - start_time
    result = {f'{split}_runtime': round(runtime, 4)}
    if num_samples is not None:
        samples_per_second = num_samples / runtime
        result[f'{split}_samples_per_second'] = round(samples_per_second, 3)
    if num_steps is not None:
        steps_per_second = num_steps / runtime
        result[f'{split}_steps_per_second'] = round(steps_per_second, 3)

    return result




import os
import logging
import json
from enum import Enum
from dataclasses import asdict, dataclass, field
from typing import Optional
from functools import cached_property
import math

import torch

from optimization import SchedulerType
from trainer_utils import IntervalStrategy


logger = logging.getLogger(__name__)


def default_logdir() -> str:
    """
    Same default as PyTorch
    """
    import socket
    from datetime import datetime

    current_time = datetime.now().strftime('%b%d_%H-%M-%S')
    # return os.path.join('runs', f'{current_time}_{socket.gethostname()}')
    return os.path.join('runs', current_time)


@dataclass
class TrainingArguments:

    output_dir: str = field(
        metadata={'help': 'The output directory where model predictions and checkpoints will be written.'},
    )

    do_train: bool = field(default=False, metadata={'help': 'Whether to run training.'})
    do_eval: bool = field(default=False, metadata={'help': 'Whether to run eval on the validation set.'})
    do_predict: bool = field(default=False, metadata={'help': 'Whether to run eval on the test set.'})

    do_early_stopping: bool = field(
        default=False, metadata={'help': 'Whether to perform early stopping during training.'}
    )
    early_stopping_patience: int = field(
        default=None, metadata={'help': 'Stop training when the specified metric worsens for X epochs.'}
    )

    trainable_loss: bool = field(
        default=False, metadata={'help': 'Whether the loss module has trainable parameters to optimize.'}
    )

    evaluation_strategy: IntervalStrategy = field(
        default='no',
        metadata={'help': 'The evaluation strategy to use.'}
    )
    eval_steps: int = field(default=None, metadata={'help': 'Run an evaluation every X steps.'})

    prediction_loss_only: bool = field(
        default=False,
        metadata={'help': 'When performing evaluation and prediction, return only the loss.'}
    )

    train_batch_size: int = field(
        default=8, metadata={'help': 'Batch size for training.'}
    )
    eval_batch_size: int = field(
        default=8, metadata={'help': 'Batch size for evaluation.'}
    )

    gradient_accumulation_steps: int = field(
        default=1,
        metadata={'help': 'Number of update steps to accumulate before performing a backward/update pass.'}
    )

    learning_rate: float = field(default=5e-4, metadata={'help': 'The initial learning rate for AdamW.'})
    weight_decay: float = field(default=0.0, metadata={'help': 'Weight decay for AdamW if applied.'})
    adam_beta1: float = field(default=0.9, metadata={'help': 'Beta1 for AdamW optimizer.'})
    adam_beta2: float = field(default=0.999, metadata={'help': 'Beta2 for AdamW optimizer.'})
    adam_epsilon: float = field(default=1e-8, metadata={'help': 'Epsilon for AdamW optimizer.'})
    max_grad_norm: float = field(default=1.0, metadata={'help': 'Max gradient norm.'})

    num_train_epochs: float = field(default=50.0, metadata={'help': 'Total number of training epochs to perform.'})
    lr_scheduler_type: SchedulerType = field(default='linear', metadata={'help': 'The scheduler type to use.'})
    warmup_ratio: float = field(
        default=0.0,
        metadata={'help': 'Linear warmup over warmup_ratio fraction of total steps.'}
    )
    warmup_steps: int = field(default=0, metadata={'help': 'Linear warmup of warmup_steps.'})

    log_level: Optional[str] = field(
        default='info',
        metadata={'help': 'Logger log level to use. Possible choices are the log levels as strings:'
                          '"debug", "info", "warning", "error", and "critical". Defaults to "info".'}
    )
    logging_strategy: IntervalStrategy = field(default='steps', metadata={'help': 'The logging strategy to use.'})
    logging_first_step: bool = field(default=False, metadata={'help': 'Log the first global_step.'})
    logging_steps: int = field(default=500, metadata={'help': 'Log every X update steps.'})
    logging_nan_inf_filter: str = field(default=True, metadata={'help': 'Filter nan and inf losses for logging.'})

    save_strategy: IntervalStrategy = field(
        default='epoch',
        metadata={'help': 'The checkpoint save strategy to use.'}
    )
    save_steps: int = field(default=500, metadata={'help': 'Save checkpoint every X update steps.'})
    save_total_limit: Optional[int] = field(
        default=None,
        metadata={'help': 'Limit the total number of checkpoints. '
                          'Deletes the older checkpoints in the output_dir. Default is unlimited checkpoints.'}
    )

    no_cuda: bool = field(default=False, metadata={'help': 'Do not use CUDA even when it is available.'})
    seed: int = field(default=42, metadata={'help': 'Random seed that will be set a the beginning of training.'})
    dataloader_drop_last: bool = field(
        default=False, metadata={'help': 'Drop the last incomplete batch if it is not divisible by the batch size.'}
    )
    dataloader_num_workers: int = field(
        default=0,
        metadata={'help': 'Number of subprocesses to use for data loading. '
                          '0 means that the data will be loaded in the main process.'}
    )
    dataloader_pin_memory: bool = field(
        default=False, metadata={'help': 'Whether or not to pin memory for DataLoader.'}
    )
    disable_tqdm: Optional[bool] = field(
        default=None, metadata={'help': 'Whether or not to disable the tqdm progress bars.'}
    )

    load_best_model_at_end: Optional[bool] = field(
        default=False,
        metadata={'help': 'Whether or not to load the best model found during training and the end of training.'}
    )
    metric_for_best_model: Optional[str] = field(
        default=None, metadata={'help': 'The metric to use to compare two different models.'}
    )
    greater_is_better: Optional[bool] = field(
        default=None, metadata={'help': 'Whether the `metric_for_best_model` should be maximized or not.'}
    )

    def __post_init__(self):
        self.log_level = logging.getLevelName(self.log_level.upper())

        if self.output_dir is not None:
            self.output_dir = os.path.expanduser(self.output_dir)
            self.logging_dir = os.path.expanduser(
                os.path.join(self.output_dir, default_logdir())
            )
            os.makedirs(self.logging_dir, exist_ok=True)

        if self.disable_tqdm is None:
            self.disable_tqdm = logger.getEffectiveLevel() > logging.WARNING

        self.evaluation_strategy = IntervalStrategy(self.evaluation_strategy)
        self.logging_strategy = IntervalStrategy(self.logging_strategy)
        self.save_strategy = IntervalStrategy(self.save_strategy)

        self.lr_scheduler_type = SchedulerType(self.lr_scheduler_type)
        if not self.do_eval and self.evaluation_strategy != IntervalStrategy.NO:
            self.do_eval = True

        # logging_steps must be non-zero for logging_strategy other than 'no'.
        if self.logging_strategy == IntervalStrategy.STEPS and self.logging_steps == 0:
            raise ValueError(f'Logging strategy {self.logging_strategy} requires non-zero `logging_steps`.')

        # Sanity checks for load_best_model_at_end: We require save and eval strategies to be compatible.
        if self.do_early_stopping:
            if self.early_stopping_patience is None:
                raise ValueError('Early stopping enabled, but no patience limit is set.')

        if self.load_best_model_at_end:
            if self.evaluation_strategy != self.save_strategy:
                raise ValueError(
                    f'`load_best_model_at_end` requires the save and eval strategy to match, but found\n'
                    f'- Evaluation strategy: {self.evaluation_strategy}\n'
                    f'- Save strategy: {self.save_strategy}'
                )

        if self.load_best_model_at_end and self.metric_for_best_model is None:
            self.metric_for_best_model = 'loss'
        if self.greater_is_better is None and self.metric_for_best_model is not None:
            self.greater_is_better = self.metric_for_best_model not in ['loss', 'eval_loss']

        if self.warmup_ratio < 0 or self.warmup_ratio > 1:
            raise ValueError('warmup_ratio must lie in the range [0, 1].')
        elif self.warmup_ratio > 0 and self.warmup_steps > 0:
            logger.info(
                f'Both warmup_ratio and warmup_steps given, '
                f'warmup_steps will override any effects of warmup_ratio during training.'
            )

    def __str__(self):
        self_as_dict = asdict(self)

        attrs_as_str = [f'{k}={v},\n' for k, v in sorted(self_as_dict.items())]
        return f"{self.__class__.__name__}(\n{''.join(attrs_as_str)})"

    __repr__ = __str__

    @cached_property
    def _setup_devices(self) -> 'torch.device':
        logger.info('PyTorch: setting up devices.')
        if self.no_cuda:
            device = 'cpu'
        else:
            device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

        if device.type == 'cuda':
            torch.cuda.set_device(device)

        return device

    @property
    def device(self) -> 'torch.device':
        return self._setup_devices

    def get_log_level(self):
        return self.log_level

    def get_warmup_steps(self, num_training_steps: int):
        warmup_steps = (
            self.warmup_steps if self.warmup_steps > 0 else math.ceil(num_training_steps * self.warmup_ratio)
        )
        return warmup_steps

    def to_dict(self):
        d = asdict(self)
        for k, v in d.items():
            if isinstance(v, Enum):
                d[k] = v.value
            if isinstance(v, list) and len(v) > 0 and isinstance(v[0], Enum):
                d[k] = [x.value for x in v]
        return d

    def to_json_string(self):
        return json.dumps(self.to_dict(), indent=2)




import os
from itertools import compress

import torch
from torch.utils.data import Dataset
import dgl


class ProvGraphDataset(Dataset):
    def __init__(self, graphs, labels=None):
        self.graphs = graphs
        self.labels = labels

    def __len__(self):
        return len(self.graphs)

    def __getitem__(self, item):
        g = self.graphs[item]
        if self.labels is not None:
            lbl = self.labels[item]
            return g, lbl
        else:
            return g
    
    def __repr__(self) -> str:
        return f'Num graphs: {len(self.graphs)} | Labels: {torch.bincount(self.labels.int())}'

    @classmethod
    def build_dataset(cls, data_dir, attack_type, multiclass=False, test_size=None, bidirected=True):
        if attack_type == 'all':
            graphs, labels = [], []
            for i, attack in enumerate(os.listdir(data_dir)):
                graphs_path = os.path.join(data_dir, attack)
                indiv_graphs, graph_dict = dgl.load_graphs(graphs_path)

                graphs.extend(indiv_graphs)

                indiv_labels = graph_dict['labels']
                if multiclass:
                    indiv_labels[indiv_labels.nonzero(as_tuple=True)] += i
                    indiv_labels = indiv_labels.long()
                else:
                    indiv_labels = indiv_labels.float()  # BCEloss expects float
                
                labels.append(indiv_labels)  # Cross entropy loss expects long-int
            
            labels = torch.cat(labels)
            
        else:
            graphs_path = os.path.join(data_dir, f'{attack_type}-graphs-new.bin')
            graphs, graph_dict = dgl.load_graphs(graphs_path)

            labels = graph_dict['labels'].float()  # BCEloss expects float
        
        if bidirected:
            graphs = [
                dgl.to_simple(dgl.to_bidirected(graph, copy_ndata=True), copy_ndata=True, return_counts=None)
                for graph in graphs
            ]

        if test_size:
            test_mask = torch.rand(len(graphs)) < test_size
            train_mask = ~test_mask

            train_dataset = cls(list(compress(graphs, train_mask)), labels[train_mask])
            test_dataset = cls(list(compress(graphs, test_mask)), labels[test_mask])

            return train_dataset, test_dataset
        else:
            return cls(graphs, labels)



from setuptools import setup

DISTNAME = 'gnn-trainer'
DESCRIPTION = 'A HuggingFace style trainer for graph neural network models'
MAINTAINER = 'Joshua Melton'
MAINTAINER_EMAIL = 'jmelto30@uncc.edu'
URL = 'https://github.com/NASCL/gnn-trainer'
LICENSE = 'MIT'

import gnn_trainer

VERSION = gnn_trainer.__version__

setup(
    name=DISTNAME,
    version=VERSION,
    description=DESCRIPTION,
    author=MAINTAINER,
    author_email=MAINTAINER_EMAIL,
    license=LICENSE,
    keywords=['pytorch', 'graph machine learning', 'graph representation learning', 'graph neural networks'],
    packages=['gnn_trainer'],
    include_package_data=True,
    install_requires=[
        'numpy>=1.16.0',
        'requests',
        'scikit-learn>=0.20.0',
        'tqdm',
        'torch>=1.6.0',
    ],
)
